/*
  Module_S88

  ** Historique:
    2020-04-20 Ulysse Delmas-Begue ajout du multiplexage (et suppression temporaire ? du filtrage)
    2020-04-20 Ulysse Delmas-Begue correction d'un pb de chainage qui ajoutait une entree fantome ente chaque module de 16 entrees
    2017-06-14 Ulysse Delmas-Begue creation

  ** License:
    GPL v2

  ** Description:

    Ce Sketch transforme un Arduino UNO/NANO en un module S88 a:
    - 16 entrees directes
    - 8  entrees directes et 16 entrées multiplexées(4x4)
    - 0  entrees directes et 16(2x8),24(3x8),32(4x8),40(5x8),48(6x8),56(7x8),64(8x8) entrées multipexées
    si la broche CLR est utilisee, reduction de 1 entree directe et mux 64 non disponible
    Dans le cas ou CLR n'est pas utilisee, la broche LOAD fait aussi office de CLR (ce qui suffit)

    Les entrees directes sont actives au niveau bas et la resistance de pull-up de l'Arduino est activee
    En cas de perturbations, vous pouvez rajouter des pull-up externes
    Les fronts sur les entrees sont memorises

    On utilise directement les ports car les acces digitalWrite/Read sont trop lent
    Le programme memorise des impulsion d'au moins 1ms sur les entrees
    et fonctionne au moins avec une horloge s88 de 10kHz

    Si PB de televersement, penser a debrancher les entrees qui sont sur TX et RX 

  ** Interface S88:
               Patte              Port
    RST      : A1 (si utilisee) : C1
    LOAD     : A2               : C2
    CLK      : A3               : C3
    Dcentrale: A4               : C4
    Dext     : A5               : C5

  ** Brochage pour la configuration 15/16 entrees directes:

    Numero        15 14 13 12   11  10  9  8                7  6  5  4  3  2  1  0
    Patte Arduino A1.A0.D13.D12.D11.D10.D9.D8               D7.D6.D5.D4.D3.D2.D1.D0
    Port  Arduino C1.C0.B5 .B4 .B3 .B2 .B1.B0    Clr        D7.D6.D5.D4.D3.D2.D1.D0
                  |  |  |   |   |   |   |  |      v         |  |  |  |  |  |  |  |
                 #O##O##O###O###O###O###O##O# Memorisation #O##O##O##O##O##O##O##O#
                  |  |  |   |   |   |   |  |                |  |  |  |  |  |  |  |
         Dext -> #O##O##O###O###O###O###O##O#### Decalage ##O##O##O##O##O##O##O##O# -> Dcentrale
                                                  ^    ^
                                                 Clk  Load

  ** Brochage pour la configuration 7/8 entrees directes et 16 entrees mux:

    D4->mux_out0 D0<-mux_in0
    D5->mux_out1 D1<-mux_in1
    D6->mux_out2 D2<-mux_in2
    D7->mux_out3 D3<-mux_in3

        in0 : entre mux_out0 et mux_in0
        in1 : entre mux_out0 et mux_in1
        in2 : entre mux_out0 et mux_in2
        in3 : entre mux_out0 et mux_in3
        in4 : entre mux_out1 et mux_in0
        ...
        in15: entre mux_out3 et mux_in3

    D8 <-in16
    D9 <-in17
    D10<-in18
    D11<-in19
    D12<-in20
    D13<-in21
    A0 <-in22
   (A1)<-in23

  ** Brochage pour les configuration 16,24,32,40,48,56,64 entrees mux

    D0<-mux_in0
    D1<-mux_in1
    D2<-mux_in2
    D3<-mux_in3
    D4<-mux_in4
    D5<-mux_in5
    D6<-mux_in6
    D7<-mux_in7

    D8->mux_out0
    D9->mux_out1
    D10->mux_out2
    D11->mux_out3
    D12->mux_out4
    D13->mux_out5
    A0->mux_out6
    A1->mux_out7

             mux_in7 mux_in6 mux_in5 mux_in4 mux_in3 mux_in2 mux_in1 mux_in0
    mux_out0   in7                                                     in0
    mux_out1   in15                                                    in8
    mux_out2
    mux_out3
    mux_out0                     ...
    mux_out1
    mux_out2
    mux_out3  in63                                                     in56                                                   

    mettre une diode pour chaque entree (cathode- cote out, anode+ cote in   out-|<|-capteur--in 

  ** Registres des ports de l'Arduino:

    Addr    Name     Bit7   Bit6   Bit5   Bit4   Bit3   Bit2   Bit1   Bit0
    ...
    0x03   PINB     PINB7  PINB6  PINB5  PINB4  PINB3  PINB2  PINB1  PINB0
    0x04   DDRB     DDB7   DDB6   DDB5   DDB4   DDB3   DDB2   DDB1   DDB0
    0x05   PORTB    PORTB7 PORTB6 PORTB5 PORTB4 PORTB3 PORTB2 PORTB1 PORTB0
    0x06   PINC     x      PINC6  PINC5  PINC4  PINC3  PINC2  PINC1  PINC0
    0x07   DDRC     x      DDC6   DDC5   DDC4   DDC3   DDC2   DDC1   DDC0
    0x08   PORTC    x      PORTC6 PORTC5 PORTC4 PORTC3 PORTC2 PORTC1 PORTC0
    0x09   PIND     PIND7  PIND6  PIND5  PIND4  PIND3  PIND2  PIND1  PIND0
    0x0A   DDRD     DDD7   DDD6   DDD5   DDD4   DDD3   DDD2   DDD1   DDD0
    0x0B   PORTD    PORTD7 PORTD6 PORTD5 PORTD4 PORTD3 PORTD2 PORTD1 PORTD0

  Mesures:

      Temps de boucle maxi: 5.25us

  Protocol S88:

      RST  ______#_________       ________       ____
      LDIN __###___________  ...  ________  ...  ____
      CLK  ___#____#___#___       _#___#__       _#__
      Data     0    1   2           15  16         31

      Le Load est synchrone pour fonctionner avec les CD4014
*/


/* configuration utilisateur */

// Configuration (mettre seulement un 1 pour la configuration selectionne)
#define use_16in  1       // 16 entrees directes
#define use_8in_16mux 0   //  8 entrees directes et 16 entrees matricees (4outx4in)
#define use_16mux 0       // 16 entrees matricees (2outx8in)
#define use_24mux 0       // 24 entrees matricees (3outx8in)
#define use_32mux 0       // 32 entrees matricees (4outx8in)
#define use_40mux 0       // 40 entrees matricees (5outx8in)
#define use_48mux 0       // 48 entrees matricees (6outx8in)
#define use_56mux 0       // 56 entrees matricees (7outx8in)
#define use_64mux 0       // 64 entrees matricees (8outx8in)

// Choisir si on utilise la broche CLR
#define use_clr_pin 0   // 1 on utilise CLR, 0 on n'utilise pas CLR alors LD fait aussi office de CLR

/* fin de la configuration utilisateur */
#if(use_16in == 1)
    #define S88_LEN 2
#elif(use_8in_16mux == 1)
    #define S88_LEN 3
#elif(use_16mux  == 1)
    #define S88_LEN 2
#elif(use_24mux  == 1)
    #define S88_LEN 3
#elif(use_32mux  == 1)
    #define S88_LEN 4
#elif(use_40mux  == 1)
    #define S88_LEN 5
#elif(use_48mux  == 1)
    #define S88_LEN 6
#elif(use_56mux  == 1)
    #define S88_LEN 7
#elif(use_64mux  == 1)
    #define S88_LEN 8  
#else
    #error "invalid configuration"
#endif

#if(use_64mux  == 1)
    #if(use_clr_pin == 1)
        #error "64mux cannot be used with CLR pin"
    #endif
#endif


byte mem0 = 0; byte mem1 = 0; byte mem2 = 0; byte mem3 = 0; byte mem4 = 0; byte mem5 = 0; byte mem6 = 0; byte mem7 = 0;
byte dec0 = 0; byte dec1 = 0; byte dec2 = 0; byte dec3 = 0; byte dec4 = 0; byte dec5 = 0; byte dec6 = 0; byte dec7 = 0;
byte portc; byte old_portc;
byte portd = 0;
byte cpt0=0;
byte cpt1=0;
byte cpt2=0;
byte cpt3=0;
byte cpt4=0;


void setup()
{
  #if(use_16in == 1)
      DDRD  = 0x00; // entrees
    //DDRD  = 0x01; // UU debug pour mesurer le temps de boucle
      PORTD = 0xff; // pull-ups
      DDRB  = 0x00; // entrees
      PORTB = 0x3f; // pull-ups
      DDRC  = 0x10; // Dcentrale est en sortie
      PORTC = 0x03; // pull-ups
  #endif
  #if(use_8in_16mux == 1)
      DDRD  = 0xf0; // 4sorties+ 4entrees
      PORTD = 0xef; // mux_out (+init D.4=0) + pull-ups
      DDRB  = 0x00; // entrees
      PORTB = 0x3f; // pull-ups
      DDRC  = 0x10; // Dcentrale est en sortie
      PORTC = 0x03; // pull-ups
  #endif
  #if((use_16mux == 1) || (use_24mux == 1) || (use_32mux == 1) || (use_40mux == 1) || (use_48mux == 1))
      DDRD  = 0x00; // entrees
      PORTD = 0xff; // pull-ups
      DDRB  = 0x3f; // sorties
      PORTB = 0x3e; // mux_out (+init B.0=0)
      DDRC  = 0x10; // Dcentrale est en sortie
      PORTC = 0x00; //
  #endif
  #if(use_56mux == 1)
      DDRD  = 0x00; // entrees
      PORTD = 0xff; // pull-ups
      DDRB  = 0x3f; // sorties
      PORTB = 0x3e; // mux_out (+init B.0=0)
      DDRC  = 0x11; // Dcentrale est en sortie et mux6_out
      PORTC = 0x01; // mux6_out a 1
  #endif
  #if(use_64mux == 1)
      DDRD  = 0x00; // entrees
      PORTD = 0xff; // pull-ups
      DDRB  = 0x3f; // sorties
      PORTB = 0x3e; // mux_out (+init B.0=0)
      DDRC  = 0x13; // Dcentrale est en sortie et mux6/7_out
      PORTC = 0x03; // mux6/7_out a 1
  #endif
  noInterrupts();  // pour ne pas etre derange
}

inline void S88(void)
{
    // decalage
    if((old_portc & 8)==0) if(portc & 8) if((portc & 4)==0) // front montant CLK sans LD: on decale
    {
        dec0 = dec0 >> 1;
        if(dec1 & 1) dec0 |= 0x80;
        dec1 = dec1 >> 1;
        #if S88_LEN == 2
        if(PINC & 0x20) dec1 |= 0x80;
        #endif

        #if S88_LEN >= 3
        if(dec2 & 1) dec1 |= 0x80;
        dec2 = dec2 >> 1;
        #if S88_LEN == 3
        if(PINC & 0x20) dec2 |= 0x80;
        #endif
        #endif

        #if S88_LEN >= 4
        if(dec3 & 1) dec2 |= 0x80;
        dec3 = dec3 >> 1;
        #if S88_LEN == 4
        if(PINC & 0x20) dec3 |= 0x80;
        #endif
        #endif

        #if S88_LEN >= 5
        if(dec4 & 1) dec3 |= 0x80;
        dec4 = dec4 >> 1;
        #if S88_LEN == 5
        if(PINC & 0x20) dec4 |= 0x80;
        #endif
        #endif

        #if S88_LEN >= 6
        if(dec5 & 1) dec4 |= 0x80;
        dec5 = dec5 >> 1;
        #if S88_LEN == 6
        if(PINC & 0x20) dec5 |= 0x80;
        #endif
        #endif

        #if S88_LEN >= 7
        if(dec6 & 1) dec5 |= 0x80;
        dec6 = dec6 >> 1;
        #if S88_LEN == 7
        if(PINC & 0x20) dec6 |= 0x80;
        #endif
        #endif

        #if S88_LEN == 8
        if(dec7 & 1) dec6 |= 0x80;
        dec7 = dec7 >> 1;
        if(PINC & 0x20) dec7 |= 0x80;
        #endif                  
    }

    // chainage
    if(old_portc & 8) if((portc & 8)==0)  // front descendant sur CLK: on met a jour Dcentrale
    {
        if(dec0 & 1) asm("sbi 8,4\n\t");  // Dcentrale=1 PATTE A4 : PORTC=8 , bit4
        else         asm("cbi 8,4\n\t");  // Dcentrale=0 PATTE A4 : PORTC=8 , bit4
    }
        
    // chargement (potentiellement avec effacement)
    if((old_portc & 8)==0) if(portc & 8) if(portc & 4) // front montant CLK avec LD: on charge
    {
        dec7 = mem7; dec6 = mem6; dec5 = mem5; dec4 = mem4; dec3 = mem3; dec2 = mem2; dec1 = mem1; dec0 = mem0;
            
        if(dec0 & 1) asm("sbi 8,4\n\t");  // Dcentrale=1 PATTE A4 : PORTC=8 , bit4
        else         asm("cbi 8,4\n\t");  // Dcentrale=0 PATTE A4 : PORTC=8 , bit4
            
        #if(use_clr_pin == 0)
            // effacement
            mem7 = mem6 = mem5 = mem4 = mem3 = mem2 = mem1 = mem0 = 0;
            cpt0 = cpt1 = cpt2 = cpt3 = cpt4 = 0;
        #endif
    }

    // effacement
    #if(use_clr_pin == 1)
        if(portc & 2)
        {
            mem7 = mem6 = mem5 = mem4 = mem3 = mem2 = mem1 = mem0 = 0;
            cpt0 = cpt1 = cpt2 = cpt3 = cpt4 = 0;
        }
    #endif
}


inline void W2US(void)  //attente 2us pour MUX = 32nop @16MHz
{
    asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t");
    asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t");
    asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t");
    asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t"); asm("nop\n\t");
}

void loop()
{
    W2US();

    while(1) // plus rapide que de sortir et re-rentrer dans loop()
    {
        // debug pour mesurer le temps de boucle
        //asm("sbi 11,0\n\t");  // PORTD=11 , bit0 // UU DBG !!
        //asm("cbi 11,0\n\t");  // PORTD=11 , bit0 // UU DBG !!

        // lecture des pins et memorisation (il faudra peut etre mettre un filtre pour eviter les glitches)
 
        #if(use_16in == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            mem1 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem1 |= 0x40;
            if((portc & 2)==0) mem1 |= 0x80;
            S88();
        #endif

        //#if(use_8in_16mux == 1)
        #if(0)  // normal use_8in_16mux
            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            // mem0 |= (PIND ^ 0xff) & 0x0f;  // possibilite de faire cela mais quasiment aussi long et on peut ne peut pas re-affecter
            if(portd & 1) mem0 |= 1;    
            if(portd & 2) mem0 |= 2;
            if(portd & 4) mem0 |= 4;
            if(portd & 8) mem0 |= 8;
            asm("sbi 11,4\n\t"); // PORTD.4=D4 @ HIGH=DIS
            asm("cbi 11,5\n\t"); // PORTD.5=D5 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem0 |= 16;
            if(portd & 2) mem0 |= 32;
            if(portd & 4) mem0 |= 64;
            if(portd & 8) mem0 |= 128;
            asm("sbi 11,5\n\t"); // PORTD.5=D5 @ HIGH=DIS
            asm("cbi 11,6\n\t"); // PORTD.6=D6 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 1;
            if(portd & 2) mem1 |= 2;
            if(portd & 4) mem1 |= 4;
            if(portd & 8) mem1 |= 8;
            asm("sbi 11,6\n\t"); // PORTD.6=D6 @ HIGH=DIS
            asm("cbi 11,7\n\t"); // PORTD.7=D7 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 16;
            if(portd & 2) mem1 |= 32;
            if(portd & 4) mem1 |= 64;
            if(portd & 8) mem1 |= 128;
            asm("sbi 11,7\n\t"); // PORTD.7=D7 @ HIGH=DIS
            asm("cbi 11,4\n\t"); // PORTD.4=D4 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();
        #endif

        #if(use_8in_16mux == 1)  // ulysse remaping
            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            // mem0 |= (PIND ^ 0xff) & 0x0f;  // possibilite de faire cela mais quasiment aussi long et on peut ne peut pas re-affecter
            if(portd & 1) mem1 |= 1;    
            if(portd & 2) mem1 |= 16;
            if(portd & 4) mem0 |= 1;
            if(portd & 8) mem0 |= 16;
            asm("sbi 11,4\n\t"); // PORTD.4=D4 @ HIGH=DIS
            asm("cbi 11,5\n\t"); // PORTD.5=D5 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 2;
            if(portd & 2) mem1 |= 32;
            if(portd & 4) mem0 |= 2;
            if(portd & 8) mem0 |= 32;
            asm("sbi 11,5\n\t"); // PORTD.5=D5 @ HIGH=DIS
            asm("cbi 11,6\n\t"); // PORTD.6=D6 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 4;
            if(portd & 2) mem1 |= 64;
            if(portd & 4) mem0 |= 4;
            if(portd & 8) mem0 |= 64;
            asm("sbi 11,6\n\t"); // PORTD.6=D6 @ HIGH=DIS
            asm("cbi 11,7\n\t"); // PORTD.7=D7 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 128;  //inv car PB opto 11
            if(portd & 2) mem1 |= 8;    //inv on remplace par le 15
            if(portd & 4) mem0 |= 8;
            if(portd & 8) mem0 |= 128;
            asm("sbi 11,7\n\t"); // PORTD.7=D7 @ HIGH=DIS
            asm("cbi 11,4\n\t"); // PORTD.4=D4 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();
        #endif

        #if 0 //(use_8in_16mux == 1)  // ulysse TEST
            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            // mem0 |= (PIND ^ 0xff) & 0x0f;  // possibilite de faire cela mais quasiment aussi long et on peut ne peut pas re-affecter
            if(portd & 1) mem1 |= 1;
            if(portd & 2) mem1 |= 16;
            if(portd & 4) { cpt0++; if(cpt0==200) mem0 |= 1;  }
            if(portd & 8) { cpt4++; if(cpt4==200) mem0 |= 16; }
            asm("sbi 11,4\n\t"); // PORTD.4=D4 @ HIGH=DIS
            asm("cbi 11,5\n\t"); // PORTD.5=D5 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 2;
            if(portd & 2) mem1 |= 32;
            if(portd & 4) { cpt1++; if(cpt1==200) mem0 |= 2; }
            if(portd & 8) mem0 |= 32;
            asm("sbi 11,5\n\t"); // PORTD.5=D5 @ HIGH=DIS
            asm("cbi 11,6\n\t"); // PORTD.6=D6 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 4;
            if(portd & 2) mem1 |= 64;
            if(portd & 4) { cpt2++; if(cpt2==200) mem0 |= 4; }
            if(portd & 8) mem0 |= 64;
            asm("sbi 11,6\n\t"); // PORTD.6=D6 @ HIGH=DIS
            asm("cbi 11,7\n\t"); // PORTD.7=D7 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            portd = PIND ^ 0xff;
            if(portd & 1) mem1 |= 8;
            if(portd & 2) mem1 |= 128;
            if(portd & 4) { cpt3++; if(cpt3==200) mem0 |= 8; }
            if(portd & 8) mem0 |= 128;
            asm("sbi 11,7\n\t"); // PORTD.7=D7 @ HIGH=DIS
            asm("cbi 11,4\n\t"); // PORTD.4=D4 @ LOW=EN
            mem2 |= (PINB ^ 0xff) & 0x3f;
            if((portc & 1)==0) mem2 |= 0x40;
            if((portc & 2)==0) mem2 |= 0x80;
            S88();
            W2US();
        #endif
        


        #if(use_16mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8 @ LOW=EN
            S88();
            W2US();
        #endif

        #if(use_24mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");
            asm("cbi 5,2\n\t");
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");
            asm("cbi 5,0\n\t");
            S88();
            W2US();            
        #endif

        #if(use_32mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,2\n\t");  // PORTB.2=D10 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem2 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");  // PORTB.2=D10 @ HIGH=DIS
            asm("cbi 5,3\n\t");  // PORTB.3=D11 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem3 |= PIND ^ 0xff;
            asm("sbi 5,3\n\t");  // PORTB.3=D11 @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8  @ LOW=EN
            S88();
            W2US();
        #endif

        #if(use_40mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,2\n\t");  // PORTB.2=D10 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem2 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");  // PORTB.2=D10 @ HIGH=DIS
            asm("cbi 5,3\n\t");  // PORTB.3=D11 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem3 |= PIND ^ 0xff;
            asm("sbi 5,3\n\t");  // PORTB.3=D11 @ HIGH=DIS
            asm("cbi 5,4\n\t");  // PORTB.4=D12 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem4 |= PIND ^ 0xff;
            asm("sbi 5,4\n\t");  // PORTB.4=D12 @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8  @ LOW=EN
            S88();
            W2US();
        #endif

        #if(use_48mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,2\n\t");  // PORTB.2=D10 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem2 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");  // PORTB.2=D10 @ HIGH=DIS
            asm("cbi 5,3\n\t");  // PORTB.3=D11 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem3 |= PIND ^ 0xff;
            asm("sbi 5,3\n\t");  // PORTB.3=D11 @ HIGH=DIS
            asm("cbi 5,4\n\t");  // PORTB.4=D12 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem4 |= PIND ^ 0xff;
            asm("sbi 5,4\n\t");  // PORTB.4=D12 @ HIGH=DIS
            asm("cbi 5,5\n\t");  // PORTB.5=D13 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem5 |= PIND ^ 0xff;
            asm("sbi 5,5\n\t");  // PORTB.5=D13 @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8  @ LOW=EN
            S88();
            W2US();
        #endif

        #if(use_56mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,2\n\t");  // PORTB.2=D10 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem2 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");  // PORTB.2=D10 @ HIGH=DIS
            asm("cbi 5,3\n\t");  // PORTB.3=D11 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem3 |= PIND ^ 0xff;
            asm("sbi 5,3\n\t");  // PORTB.3=D11 @ HIGH=DIS
            asm("cbi 5,4\n\t");  // PORTB.4=D12 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem4 |= PIND ^ 0xff;
            asm("sbi 5,4\n\t");  // PORTB.4=D12 @ HIGH=DIS
            asm("cbi 5,5\n\t");  // PORTB.5=D13 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem5 |= PIND ^ 0xff;
            asm("sbi 5,5\n\t");  // PORTB.5=D13 @ HIGH=DIS
            asm("cbi 8,0\n\t");  // PORTC.0=D8  @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem6 |= PIND ^ 0xff;
            asm("sbi 8,0\n\t");  // PORTC.0=A0  @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8  @ LOW=EN
            S88();
            W2US();            
        #endif

        #if(use_64mux == 1)
            old_portc = portc;
            portc = PINC;
            mem0 |= PIND ^ 0xff;
            asm("sbi 5,0\n\t");  // PORTB.0=D8 @ HIGH=DIS
            asm("cbi 5,1\n\t");  // PORTB.1=D9 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem1 |= PIND ^ 0xff;
            asm("sbi 5,1\n\t");  // PORTB.1=D9  @ HIGH=DIS
            asm("cbi 5,2\n\t");  // PORTB.2=D10 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem2 |= PIND ^ 0xff;
            asm("sbi 5,2\n\t");  // PORTB.2=D10 @ HIGH=DIS
            asm("cbi 5,3\n\t");  // PORTB.3=D11 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem3 |= PIND ^ 0xff;
            asm("sbi 5,3\n\t");  // PORTB.3=D11 @ HIGH=DIS
            asm("cbi 5,4\n\t");  // PORTB.4=D12 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem4 |= PIND ^ 0xff;
            asm("sbi 5,4\n\t");  // PORTB.4=D12 @ HIGH=DIS
            asm("cbi 5,5\n\t");  // PORTB.5=D13 @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem5 |= PIND ^ 0xff;
            asm("sbi 5,5\n\t");  // PORTB.5=D13 @ HIGH=DIS
            asm("cbi 8,0\n\t");  // PORTC.0=D8  @ LOW=EN
            S88();
            W2US();

            old_portc = portc;
            portc = PINC;
            mem6 |= PIND ^ 0xff;
            asm("sbi 8,0\n\t");  // PORTC.0=A0  @ HIGH=DIS
            asm("cbi 8,1\n\t");  // PORTC.1=A1  @ LOW=EN
            S88();
            W2US();            

            old_portc = portc;
            portc = PINC;
            mem7 |= PIND ^ 0xff;
            asm("sbi 8,1\n\t");  // PORTC.1=A1  @ HIGH=DIS
            asm("cbi 5,0\n\t");  // PORTB.0=D8  @ LOW=EN
            S88();
            W2US();            
        #endif

    }
}
